/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __DAEDALUS_D3DRENDER_H__
#define __DAEDALUS_D3DRENDER_H__

#include "DLParser.h"		// DL_PF

#include "Utility\Unique.h"

//#include "DBGConsole.h"
//#include "Memory.h"

#define CLIPPING 0

#define DAEDALUS_MATRIX_STACK		60

//
// For VRML scene export
// 
//#define VRML_EXPORT_SUPPORT


#define FVF_DaedalusVtx (D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 |D3DFVF_TEX2) 
typedef struct
{
	D3DXVECTOR3 pos;
	float rhw;
	DWORD  color;
	D3DXVECTOR2  t0;
	D3DXVECTOR2  t1;
} DaedalusVtx;


/*
// This is no longer used - we specify texture coords for all vertices, even if
// they are not used. This lets us use a single vertex buffer, which is much easier to manage.
#define DAEDALUSFVF_LITVERTEX (D3DFVF_XYZRHW | D3DFVF_DIFFUSE ) 
 typedef struct _LITVERTEX {
	D3DXVECTOR3 pos;
	float rhw;
	DWORD  dcDiffuse;
} LITVERTEX, *LPLITVERTEX;*/

typedef struct
{
	D3DXVECTOR3 dir;		// Should be normalised
	D3DXCOLOR col;

} DaedalusLight;


// Bits for clipping
// +-+-+-
// xxyyzz
#define Z_NEG  0x01
#define Z_POS  0x02
#define Y_NEG  0x04
#define Y_POS  0x08
#define X_NEG  0x10
#define X_POS  0x20

#ifndef RGBA_GETRED
#define RGBA_GETALPHA(x)  (((x)>>24)& 0xff)
#define RGBA_GETRED(x)  (((x)>>16)& 0xff)
#define RGBA_GETGREEN(x)  (((x)>>8)& 0xff)
#define RGBA_GETBLUE(x)  (((x))& 0xff)
#endif

enum
{
	D3DRENDER_LOAD = 0,
	D3DRENDER_MUL = 1
};


#pragma pack(1)
typedef struct _FiddledVtx
{
	short y;
	short x;

	short flag;
	short z;

	short tv;
	short tu;

	union {
		struct _rgba {
			BYTE a;
			BYTE b;
			BYTE g;
			BYTE r;
		} rgba;
		struct _norma {
			char na;
			char nz;	// b
			char ny;	// g
			char nx;	// r
		} norma;
	};
} FiddledVtx;

typedef struct _FiddledVtxDKR
{
	s16 y;
	s16 x;

	u8 a;
	u8 b;
	s16 z;

	u8 g;
	u8 r;

} FiddledVtxDKR;

#pragma pack()


#ifdef VRML_EXPORT_SUPPORT
class SceneObject
{
	protected:
		std::vector< DaedalusVtx > verts;
		CHAR szTextureURL[ MAX_PATH + 1 ];

	public:
		void AddVertex( const DaedalusVtx & v )
		{
			verts.push_back( v );
		}

		void SetTextureURL( const CHAR * szFileName )
		{
			strcpy( szTextureURL, szFileName );
		}

		const CHAR * GetTextureURL( ) const
		{
			return szTextureURL;
		}

		u32 GetNumVerts() const
		{
			return verts.size();
		}

		D3DXVECTOR3 GetPos( u32 i ) const
		{
			return verts[i].pos;
		}

		DWORD GetColor( u32 i ) const
		{
			return verts[i].color;
		}

		D3DXVECTOR3 GetTextureCoords( u32 i ) const
		{
			return verts[i].t;
		}
};
#endif // VRML_EXPORT_SUPPORT

//*****************************************************************************
//
//*****************************************************************************
class D3DRender : public CUnique< D3DRender >
{
protected:
	friend class CUnique< D3DRender >;
	D3DRender();

public:
	~D3DRender();

	HRESULT InvalidateDeviceObjects()
	{
		for ( u32 i = 0; i < 2; i++ )
		{
			if ( m_lpsTexture[i] != NULL )
			{
				m_lpsTexture[i]->Release();
				m_lpsTexture[i] = NULL;
			}
		}

		if ( m_lpVtxBuffer )
		{
			m_lpVtxBuffer->Release();
			m_lpVtxBuffer = NULL;
		}

		if ( m_lpIdxBuffer )
		{
			m_lpIdxBuffer->Release();
			m_lpIdxBuffer = NULL;
		}
		return S_OK;
	}

	HRESULT RestoreDeviceObjects();

	void SetTextureEnable(BOOL bEnable)
	{
		m_bTextureEnable = bEnable;
	}

#ifdef VRML_EXPORT_SUPPORT
	void StartSceneDump()
	{
		m_arrObjects.clear();
		m_DumpingScene = true;
	}

	void EndSceneDump();
#endif // VRML_EXPORT_SUPPORT

	void SetAddressU( u32 index, DWORD dwMode );
	void SetAddressV( u32 index, DWORD dwMode );

	void SetAmbientLight( float r, float g, float b)
	{
		m_AmbientCol.r = r * 255.0f;
		m_AmbientCol.g = g * 255.0f;
		m_AmbientCol.b = b * 255.0f;
	}

	void	SetPrimitiveColor(DWORD dwCol)			{ m_dwPrimitiveColor = dwCol; }
	void	SetEnvColor(DWORD dwCol)				{ m_dwEnvColor = dwCol; }
	void	SetLighting(BOOL bLighting)				{ m_bLightingEnable = bLighting; }
	void	ZBufferEnable(BOOL bZBuffer)			{ m_bZBuffer = bZBuffer; }

	HRESULT FlushTris();
	DWORD	GetNumTrisRendered() const				{ return m_dwNumTrisRendered; }
	DWORD	GetNumTrisClipped() const				{ return m_dwNumTrisClipped; }

	void	SetFogMult( float fFogMult )			{ m_fFogMult = fFogMult; }
	void	SetFogOffset( float fFogOffset )		{ m_fFogOffset = fFogOffset; }
	void	SetFogEnable(BOOL bEnable)				{ m_bFogEnable = bEnable; }
	void	SetFogColor(DWORD dwCol)				{ m_dwFogColor = dwCol; }

	void	SetTexture( u32 index, LPDIRECT3DTEXTURE8 lpsTexture, LONG  nTileLeft, LONG nTileTop, DWORD dwTileWidth, DWORD dwTileHeight);
	void	SetTextureGen(BOOL bTextureGen);
	void	SetTextureScale(float fScaleX, float fScaleY);

	void	SetCullMode( BOOL bCullFront, BOOL bCullBack )
	{
		m_bCullFront = bCullFront;
		m_bCullBack = bCullBack;
	}

	void	SetSmooth( bool bSmooth )				{ mSmooth = bSmooth; }
	void	SetSmoothShade( bool bSmoothShade )		{ mSmoothShade = bSmoothShade; }
	
	void SetAlphaRef(DWORD dwAlpha)
	{
		if (m_dwAlpha != dwAlpha)
		{
			m_dwAlpha = dwAlpha;
			g_pD3DDev->SetRenderState(D3DRS_ALPHAREF, dwAlpha);
		}
	}

	enum CycleType
	{
		CYCLE_1CYCLE = 0,		// Please keep in this order - matches RDP 
		CYCLE_2CYCLE,
		CYCLE_COPY,
		CYCLE_FILL,
	};



	// Assumes dwAddress has already been checked!	
	void SetNewVertexInfo(DWORD dwAddress, DWORD dwV0, DWORD dwNum);

	// Assumes dwAddress has already been checked!	
	// DKR seems to use longer vert info
	void SetNewVertexInfoDKR(DWORD dwAddress, DWORD dwV0, DWORD dwNum);

	void SetVtxColor( u32 vert, u32 color )
	{
		if ( vert < MAX_VERTS )
		{
			m_VtxTransformed[vert].color = color;
		}
	}

	void SetVtxTextureCoord( u32 vert, short tu, short tv )
	{
		if ( vert < MAX_VERTS )
		{
			m_VtxTransformed[vert].t0.x = (float)tu / 32.0f;
			m_VtxTransformed[vert].t0.y = (float)tv / 32.0f;
			m_VtxTransformed[vert].t1 = m_VtxTransformed[vert].t0;
		}
	}

	void SetVtxXY( u32 vert, float x, float y )
	{
		if ( vert < MAX_VERTS )
		{
			// Convert to screen coords here
			m_VtxTransformed[vert].pos.x = /*round*/( x * m_fScreenMultX );
			m_VtxTransformed[vert].pos.y = /*round*/( y * m_fScreenMultY );

			//m_VtxTransformed[vert].pos.z = (vecProjected.z + 1.0f) * 0.5f;
		}
	}


	HRESULT TexRect( const CVector2 & xy0, const CVector2 & xy1, const CVector2 & uv0, const CVector2 & uv1 );
	HRESULT TexRectFlip( const CVector2 & xy0, const CVector2 & xy1, const CVector2 & uv0, const CVector2 & uv1 );
	HRESULT FillRect( const CVector2 & xy0, const CVector2 & xy1, u32 color );

	void SetProjection(const D3DXMATRIX & mat, BOOL bPush, LONG nLoadReplace);
	void SetWorldView(const D3DXMATRIX & mat, BOOL bPush, LONG nLoadReplace);
	void PopProjection();
	void PopWorldView();

	void SetScreenSize( u32 width, u32 height )
	{
		m_fScreenMultX = (float)width / (float)mViWidth;
		m_fScreenMultY = (float)height / (float)mViHeight;
	}


	void SetN64Viewport( const CVector3 & scale, const CVector3 & trans )
	{
		mVpScale = scale;
		mVpTrans = trans;

		mVpScale.x *= (float)m_fScreenMultX;
		mVpScale.y *= -(float)m_fScreenMultY;		// Apply the flip here to invert the y to the correct value
		mVpScale.z *= 1.0f / 256.0f;

		mVpTrans.x *= (float)m_fScreenMultX;
		mVpTrans.y *= (float)m_fScreenMultY;
		mVpTrans.z *= 1.0f / 256.0f;
	}
	
	void	SetScissor( s32 x0, s32 y0, s32 x1, s32 y1 )
	{
		D3DRECT rect;
		rect.x1 = s32(x0 * m_fScreenMultX);
		rect.y1 = s32(y0 * m_fScreenMultY);
		rect.x2 = s32(x1 * m_fScreenMultX);
		rect.y2 = s32(y1 * m_fScreenMultY);

#ifdef _XBOX
		g_pD3DDev->SetScissors( 1, FALSE, &rect );
#endif
	}
		
	// Returns true if triangle visible and rendered, false otherwise
	bool AddTri(u32 v0, u32 v1, u32 v2)
	{
		if ( TestTri( v0, v1, v2 ) )
		{
			DL_PF("   Tri: %d,%d,%d", v0, v1, v2);

			m_swIndexBuffer[ m_dwNumIndices + 0 ] = (u16)v0;
			m_swIndexBuffer[ m_dwNumIndices + 1 ] = (u16)v1;
			m_swIndexBuffer[ m_dwNumIndices + 2 ] = (u16)v2;

			m_dwNumIndices += 3;

			m_dwNumTrisRendered++;
			return true;
		}
		else
		{
			DL_PF("   Tri: %d,%d,%d (clipped)", v0, v1, v2);

			m_dwNumTrisClipped++;
			return false;
		}
	}

	

	void SetLightCol(DWORD dwLight, DWORD dwCol)
	{
		m_Lights[dwLight].col.r = (float)((dwCol >> 24)&0xFF);
		m_Lights[dwLight].col.g = (float)((dwCol >> 16)&0xFF);
		m_Lights[dwLight].col.b = (float)((dwCol >>  8)&0xFF);
		m_Lights[dwLight].col.a = 255.0f;	// Ignore light alpha
	}

	void SetLightDirection(DWORD dwLight, float x, float y, float z)
	{
		// Invert the x/y/z
		m_Lights[dwLight].dir.x = x;
		m_Lights[dwLight].dir.y = y;
		m_Lights[dwLight].dir.z = z;

		D3DXVec3Normalize(&m_Lights[dwLight].dir,
						  &m_Lights[dwLight].dir);

	}

	void SetNumLights(DWORD dwNumLights)
	{
		m_dwNumLights = dwNumLights;
	}

public:
	// Init matrix stack to identity matrices
	void ResetMatrices()
	{
		D3DXMATRIX mat;

		D3DXMatrixIdentity(&mat);

		m_dwProjectionTop = 0;
		m_dwModelViewTop = 0;
		m_mProjection[0] = mat;
		m_mModelView[0] = mat;
	}

#ifndef DAEDALUS_RELEASE_BUILD
	void PrintActive()
	{
		if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
		{
			D3DXMATRIX mat = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];
			DL_PF(
				" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
				" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
				" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
				" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
				mat.m[0][0], mat.m[0][1], mat.m[0][2], mat.m[0][3],
				mat.m[1][0], mat.m[1][1], mat.m[1][2], mat.m[1][3],
				mat.m[2][0], mat.m[2][1], mat.m[2][2], mat.m[2][3],
				mat.m[3][0], mat.m[3][1], mat.m[3][2], mat.m[3][3]);
		}
	}
#endif

	bool TestVerts( u32 v0, u32 vn ) const;				// Returns true if bounding volume is visible, false if culled

protected:
#ifdef USE_PIXEL_SHADERS
	void	InitCycle1PixelShader();
	void	InitCycle2PixelShader();
	void	InitCycle2FogShadePixelShader();
#else
	void	InitCycle12( bool two_cycles );
	HRESULT RenderWaterHack( DWORD vtxbase, DWORD idxbase );
	HRESULT RenderNSHack( DWORD vtxbase, DWORD idxbase );
#endif

	void	InitCycleCopy( );
	void	InitCycleFill( );
	void	InitCycle1( );
	void	InitCycle2( );


	void SetTextureFactor(DWORD dwCol);

	bool TestTri(u32 v0, u32 v1, u32 v2) const;			// Returns true if ok, false if clipped
	DWORD LightVert( const D3DXVECTOR3 & norm );

	void InitBlendMode();
	HRESULT InitialiseBuffers( DWORD * pvtxbase, DWORD * pidxbase, const DaedalusVtx *p_src_verts );


	#ifdef VRML_EXPORT_SUPPORT
	void DumpVerticesToSceneFile();
	#endif

	HRESULT	RenderTriangleList( const DaedalusVtx * p_verts, u32 num_verts, const u16 * p_indices, u32 num_indices, bool disable_zbuffer = false, bool force_point_filter = false );

protected:
	enum { MAX_VERTS = 80 };		// F3DLP.Rej supports up to 80 verts!

	u32		mViWidth;
	u32		mViHeight;

	float m_fScreenMultX;
	float m_fScreenMultY;

	CVector3	mVpScale;
	CVector3	mVpTrans;

	BOOL m_bTextureEnable;
	BOOL m_bLightingEnable;
	
	D3DCOLORVALUE m_AmbientCol;

	LONG  m_dwNumLights;
	DWORD m_dwTextureFactor;

	BOOL m_bZBuffer;

	BOOL m_bCullFront;
	BOOL m_bCullBack;
	
	DWORD m_dwAlpha;

	bool	mSmooth;
	bool	mSmoothShade;

	float m_fFogMult;
	float m_fFogOffset;

	BOOL  m_bFogEnable;
	DWORD m_dwFogColor;
	DWORD m_dwPrimitiveColor;
	DWORD m_dwEnvColor;

	// Texutring
	LPDIRECT3DTEXTURE8 m_lpsTexture[2];
	DaedalusLight m_Lights[8];

	float m_fTileWidth[2];		// Float to avoid converts when processing verts
	float m_fTileHeight[2];
	float m_fTexWidth[2];
	float m_fTexHeight[2];		// Float to avoid converts when processing verts

	// Reciprocals for fast mults
	float m_fInvTileWidth[2];
	float m_fInvTileHeight[2];
	float m_fInvTexWidth[2];
	float m_fInvTexHeight[2];

	float m_fTexLeft[2];
	float m_fTexTop[2];
	float m_fTexScaleX;
	float m_fTexScaleY;

	BOOL m_bTextureGen;			// Generate texture coords?

	D3DXMATRIX m_mProjection[DAEDALUS_MATRIX_STACK];
	D3DXMATRIX m_mModelView[DAEDALUS_MATRIX_STACK];
	DWORD	  m_dwProjectionTop;
	DWORD	  m_dwModelViewTop;
		
	// Software vertex buffer - verts are cached here and written to 
	// D3DVB in one go.
	WORD m_swIndexBuffer[500];
	DWORD m_dwNumVertices;
	DWORD m_dwNumIndices;

	// BEGIN ADDED BY Lkb - 8/jun/2001 - New combiner
	BYTE m_ucTempVertexBuffer[1000*32];
	// END ADDED BY Lkb - 8/jun/2001

	// D3D vertex buffer
	LPDIRECT3DVERTEXBUFFER8 m_lpVtxBuffer;
	LPDIRECT3DINDEXBUFFER8 m_lpIdxBuffer;
	DWORD m_dwVtxBufBase;
	DWORD m_dwVtxBufSize;

	DWORD m_dwIdxBufBase;
	DWORD m_dwIdxBufSize;

	//
	// Stats
	//
	DWORD m_dwNumTrisRendered;
	DWORD m_dwNumTrisClipped;

public:			// For DL_PF

#ifdef VRML_EXPORT_SUPPORT
	D3DXVECTOR3 m_vecPreProjected[MAX_VERTS];

	DaedalusVtx m_SceneVerts[ 10000 ];	// Temporary cache
	std::vector< SceneObject > m_arrObjects;
	bool	m_DumpingScene;

#endif



	// Processed vertices waiting for output...
	DaedalusVtx m_VtxTransformed[MAX_VERTS];

	DWORD		m_dwVecFlags[MAX_VERTS];			// Z_POS Z_NEG etc

	// Reset for a new frame
	void Reset();

	HRESULT		BeginScene();
	HRESULT		EndScene();
	void		DisplayStats();

};

#endif // __DAEDALUS_D3DRENDER_H__
