/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
#ifdef _MSC_VER
#include <crtdbg.h>
#endif
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support

#include "Resources\resource.h"			// IDS_UNSUPPORTEDTEXFMT
#include "Texture.h"
#include "PngUtil.h"

#include "Debug\DBGConsole.h"

// Probably shouldn't need more than 4096 * 4096
static u8 g_ucTempBuffer[1024*1024];

//*************************************************************************************
// Macros
//*************************************************************************************
// Set this to non-zero to force a maximum texture size (to test scaling)
#define MAX_TEXTURE_DIM		0

//*****************************************************************************
//
//*****************************************************************************
class ITexture : public CTexture
{
	public:
		ITexture();
		virtual ~ITexture();

		HRESULT Initialise( u32 width, u32 height );

#ifdef LIBPNG_SUPPORT
		BOOL DumpImageAsPNG(LPCTSTR szFileName, u32 width, u32 height );
#endif

		LPDIRECT3DTEXTURE8 GetTexture()					{ return mpTexture; }
		u32			GetPixelSize() const				{ return mPixelSize; }

		D3DFORMAT	GetSurfaceFormat() const			{ return mSurfaceFormat; }
		void		SetSurfaceFormat( D3DFORMAT fmt );
		
		// Provides access to "surface"
		BOOL StartUpdate(DrawInfo *di);
		void EndUpdate(DrawInfo *di);

	protected:

		HRESULT CreateTexture( u32 width, u32 height, LPDIRECT3DTEXTURE8 * ppSurf );
		void ScaleTempBufferToSurface();
		void CopyTempBufferToSurface();
#ifdef _XBOX
		void SwizzleTempBufferToSurface();
#endif

	protected:

		LPDIRECT3DTEXTURE8		mpTexture;
		
		u32						mWidth;			// The requested Texture w/h
		u32						mHeight;

		u32						mCorrectedWidth;	// What was actually created
		u32						mCorrectedHeight;

		u32						mPixelSize;			// Size of each pixel, in bytes
		D3DFORMAT				mSurfaceFormat;		// D3D Surface format of this texture
};

//*****************************************************************************
// Get the next power of 2 higher that the specified number
//*****************************************************************************
u32 GetNextPowerOf2( u32 x )
{
	u32 n = 1;
	while ( n < x )
	{
		n = n<<1;
	}

	return n;
}

//*****************************************************************************
//
//*****************************************************************************
CTexture * CTexture::Create( u32 width, u32 height )
{
	HRESULT hr;
	ITexture * pTexture = new ITexture();

	hr = pTexture->Initialise( width, height );
	if (FAILED(hr))
	{
		delete pTexture;
		return NULL;
	}

	return pTexture;
}



//*****************************************************************************
//
//*****************************************************************************
ITexture::ITexture() :
	mpTexture(NULL),
	mWidth(0),
	mHeight(0),
	mCorrectedWidth(0),
	mCorrectedHeight(0),
	mSurfaceFormat( D3DFMT_A8R8G8B8 ),
	mPixelSize( 4 )
{
}


//*****************************************************************************
//
//*****************************************************************************
ITexture::~ITexture()
{
	if (mpTexture)
	{
		mpTexture->Release();
		mpTexture = NULL;
	}
}

//*****************************************************************************
// Resize the DIB. Just passes control to
// CreateDIB()
//*****************************************************************************
HRESULT ITexture::Initialise( u32 width, u32 height )
{
	HRESULT hr;

	_ASSERTE(mpTexture == NULL);
	
	hr = CreateTexture( width, height, &mpTexture );
	if (FAILED(hr))
		return hr;

	mWidth = width;
	mHeight = height;
	return S_OK;

}

//*****************************************************************************
// Create the surface. Destroys old surface if
// necessary
//*****************************************************************************
HRESULT ITexture::CreateTexture( u32 width, u32 height, LPDIRECT3DTEXTURE8 * ppSurf )
{
	HRESULT hr;

#define FORMAT_32	D3DFMT_A8R8G8B8		//D3DFMT_LIN_A8B8G8R8
#define FORMAT_16	D3DFMT_A4R4G4B4		//D3DFMT_LIN_A4R4G4B4

	D3DFORMAT pf = FORMAT_32;
	UINT mip_levels = 1;//D3DX_DEFAULT;

	UINT corrected_width = GetNextPowerOf2( width );
	UINT corrected_height = GetNextPowerOf2( height );

	hr = D3DXCheckTextureRequirements(g_pD3DDev,
							&corrected_width, &corrected_height,
							&mip_levels,
							0, //
							&pf,
							D3DPOOL_MANAGED);
	if ( FAILED(hr) || ( pf != FORMAT_32 && pf != FORMAT_16 ) )
	{
		// Try again with the other supported format
		pf = FORMAT_16;
		mip_levels = 0;//D3DX_DEFAULT;

		corrected_width = GetNextPowerOf2( width );
		corrected_height = GetNextPowerOf2( height );

		hr = D3DXCheckTextureRequirements(g_pD3DDev,
								&corrected_width, &corrected_height,
								&mip_levels, // MipLevels
								0, //
								&pf,
								D3DPOOL_MANAGED);
	}

#if MAX_TEXTURE_DIM
	if ( width > MAX_TEXTURE_DIM ) corrected_width = MAX_TEXTURE_DIM;
	if ( height > MAX_TEXTURE_DIM ) corrected_height = MAX_TEXTURE_DIM;
#endif

	if (SUCCEEDED(hr))
	{
		hr = D3DXCreateTexture(g_pD3DDev,
								corrected_width, corrected_height,
								1,//D3DX_DEFAULT, // MipLevels
								0, //
								pf,
								D3DPOOL_MANAGED,
								ppSurf);
	}

	mCorrectedWidth  = corrected_width;
	mCorrectedHeight = corrected_height;

	SetSurfaceFormat( pf );

	// D3D likes texture w/h to be a power of two, so the condition below
	// will almost always hold. 
	// D3D should usually create textures large enough (on nVidia cards anyway),
	// and so there will usually be some "slack" left in the texture (blank space
	// that isn't used). The D3DRender code takes care of this (mostly)
	// Voodoo cards are limited to 256x256 and so often textures will be
	// created that are too small for the required dimensions. We compensate for
	// this by passing in a dummy area of memory when the surface is locked,
	// and copying across pixels to the real surface when the surface is unlocked.
	// In this case there will be no slack and D3DRender takes this into account.
	// 
	if (FAILED(hr))
	{
		char szError[200+1];
		D3DXGetErrorString(hr, szError, 200);
		DBGConsole_Msg(0, "!!Unable to create surface!! %d x %d (%s)", width, height, szError);
		return hr;
	}
	
	return S_OK;		
}

//*****************************************************************************
// Set the format of this texture
//*****************************************************************************
void	ITexture::SetSurfaceFormat( D3DFORMAT fmt )
{
	mSurfaceFormat = fmt;

	if ( fmt == FORMAT_32 )
	{
		mPixelSize = 4;	
	}
	else if ( fmt ==  FORMAT_16)
	{
		mPixelSize = 2;
	}
	else
	{
		DAEDALUS_ASSERTMSG( "Unknown pixel format" );

		static bool warned = false;
		if (!warned)
		{
			TCHAR szMsg[300+1];

			// IDS_UNSUPPORTEDTEXFMT takes one %d arg
			wsprintf( szMsg, "Unsupported Texture Format\n%d", fmt );
			MessageBox(NULL, szMsg, g_szDaedalusName, MB_ICONEXCLAMATION|MB_OK);
			warned = true;
		}
	}

}


//*****************************************************************************
// Get information about the DIBitmap
// This locks the bitmap (and stops 
// it from being resized). Must be matched by a
// call to EndUpdate();
//*****************************************************************************
BOOL ITexture::StartUpdate(DrawInfo *di)
{
	if (mpTexture == NULL)
		return FALSE;

	// Return a temporary buffer to use
	di->Surface = g_ucTempBuffer;
#ifdef _XBOX
	// Swizzling is faster if the pitch matches the dest texture
	di->Width   = mCorrectedWidth;
	di->Height  = mCorrectedHeight;
	di->Pitch   = mCorrectedWidth * 4;		// We always decompress to 32bpp
#else
	di->Width   = mWidth;
	di->Height  = mHeight;
	di->Pitch   = mWidth * 4;				// We always decompress to 32bpp
#endif

	return TRUE;
}

//*****************************************************************************
// This releases the DIB information, allowing it
// to be resized again
//*****************************************************************************
void ITexture::EndUpdate(DrawInfo *di)
{
	if (mpTexture == NULL)
		return;

#ifdef DAEDALUS_GFX_PLUGIN

	if (mCorrectedWidth < mWidth ||
		mCorrectedHeight < mHeight)
	{
		ScaleTempBufferToSurface();
	}
	else
	{
		CopyTempBufferToSurface();
	}

#else
	SwizzleTempBufferToSurface();
#endif
}

#ifdef _XBOX
//*************************************************************************************
// Copy and swizzle data from the temporary surface to the texture
//*************************************************************************************
void	ITexture::SwizzleTempBufferToSurface()
{
	D3DLOCKED_RECT d3d_lr;
	HRESULT hr;

	hr = mpTexture->LockRect(0, &d3d_lr, NULL, 0);
	if (SUCCEEDED(hr))
	{
		XGSwizzleRect( g_ucTempBuffer, 0, NULL, d3d_lr.pBits,
					  mCorrectedWidth, mCorrectedHeight, 
					  NULL, GetPixelSize() );

		mpTexture->UnlockRect( 0 );

	}
}
#endif

//*****************************************************************************
// Copy data from the temporary surface to the texture
//*****************************************************************************
void	ITexture::CopyTempBufferToSurface()
{
	HRESULT hr;
	D3DLOCKED_RECT d3d_lr;
	
	hr = mpTexture->LockRect(0, &d3d_lr, NULL, D3DLOCK_NOSYSLOCK);
	
	if (SUCCEEDED(hr))
	{
		if ( GetSurfaceFormat() == FORMAT_32 )
		{
			u32 * p_src = (u32*)g_ucTempBuffer;

			for ( u32 y = 0; y < mHeight; y++ )
			{
				u32 * p_dst = (u32*)( (u8*)d3d_lr.pBits + y * d3d_lr.Pitch );
				
				for ( u32 x = 0; x < mWidth; x++ )
				{
					*p_dst++ = *p_src++;
				}
			}
		}
		else if ( GetSurfaceFormat() == FORMAT_16 )
		{
			u32 * p_src = (u32*)g_ucTempBuffer;

			for ( u32 y = 0; y < mHeight; y++ )
			{
				u16 * p_dst = (u16*)( (u8*)d3d_lr.pBits + y * d3d_lr.Pitch );
				
				for ( u32 x = 0; x < mWidth; x++ )
				{
					u32 s_col = *p_src++;

					// aaaa0000rrrr0000gggg0000bbbb0000
					//                 aaaarrrrggggbbbb 

					u16 a = u16( s_col >> 16 ) & 0xf000;
					u16 r = u16( s_col >> 12 ) & 0x0f00;
					u16 g = u16( s_col >>  8 ) & 0x00f0;
					u16 b = u16( s_col >>  4 ) & 0x000f;

					u16 d_col = a | r | g | b;

					*p_dst++ = d_col;
				}
			}
		}
		else
		{
			DAEDALUS_ASSERTMSG( "Unknown pixel format" );
		}
			
		mpTexture->UnlockRect( 0 );
	}
}


//*****************************************************************************
//
//*****************************************************************************
void ITexture::ScaleTempBufferToSurface()
{
	HRESULT hr;
	u32 xDst, yDst;
	u32 xSrc, ySrc;
	D3DLOCKED_RECT d3d_lr;
	
	if (mpTexture == NULL)
		return;

	hr = mpTexture->LockRect(0, &d3d_lr, NULL, D3DLOCK_NOSYSLOCK);

	if ( SUCCEEDED(hr) )
	{
		// Copy across from the temp buffer to the surface
		if ( GetSurfaceFormat() == FORMAT_32 )
		{
			u32 * pDst;
			u32 * pSrc;

			u8 * p_src_base = (u8*)g_ucTempBuffer;
			u8 * p_dst_base = (u8*)d3d_lr.pBits;

			s32 src_pitch = mWidth * 4;
			s32 dst_pitch = d3d_lr.Pitch;

			for (yDst = 0; yDst < mCorrectedHeight; yDst++)
			{
				// ySrc ranges from 0..mHeight
				// I'd rather do this but sometimes very narrow (i.e. 1 pixel)
				// surfaces are created which results in  /0...
				//ySrc = (yDst * (mHeight-1)) / (mCorrectedHeight-1);
				ySrc = (yDst * (mHeight)) / (mCorrectedHeight);

				pSrc = (u32*)(p_src_base + (ySrc * src_pitch));
				pDst = (u32*)(p_dst_base + (yDst * dst_pitch));

				for (xDst = 0; xDst < mCorrectedWidth; xDst++)
				{
					//xSrc = (xDst * (mWidth-1)) / (mCorrectedWidth-1);
					xSrc = (xDst * (mWidth)) / (mCorrectedWidth);
					pDst[xDst] = pSrc[xSrc];
				}
			}
		}
		else if ( GetSurfaceFormat() == FORMAT_16 )
		{
			u16 * pDst;
			u16 * pSrc;

			u8 * p_src_base = (u8*)g_ucTempBuffer;
			u8 * p_dst_base = (u8*)d3d_lr.pBits;

			s32 src_pitch = mWidth * 2;
			s32 dst_pitch = d3d_lr.Pitch;

			for (yDst = 0; yDst < mCorrectedHeight; yDst++)
			{
				// ySrc ranges from 0..mHeight
				ySrc = (yDst * (mHeight)) / (mCorrectedHeight);

				pSrc = (u16*)(p_src_base + (ySrc * src_pitch));
				pDst = (u16*)(p_dst_base + (yDst * dst_pitch));

				for (xDst = 0; xDst < mCorrectedWidth; xDst++)
				{
					xSrc = (xDst * (mWidth)) / (mCorrectedWidth);
					pDst[xDst] = pSrc[xSrc];
				}
			}
		}

		mpTexture->UnlockRect( 0 );
	}
}

#ifdef LIBPNG_SUPPORT
//*****************************************************************************
//
//*****************************************************************************
BOOL ITexture::DumpImageAsPNG( LPCTSTR szFileName,  u32 width, u32 height )
{
	DrawInfo di;

	if ( StartUpdate( &di ) )
	{
		u32 *		p_dst = new u32 [ width * height ];
		u8 *		p_src_base = (u8 *)di.Surface;
		s32			pitch = di.Pitch;
		u32			index = 0;

		for ( u32 y = 0; y < height; y++ )
		{
			u32 * p_src = (u32 *)( p_src_base + y * pitch );

			memcpy( &p_dst[ index ], p_src, width * 4 );

			index += width;
		}
		EndUpdate(&di);
		
		PngSaveImage(szFileName, (u8 *)p_dst, width, height);

		delete [] p_dst;

		return TRUE;
	}
	else
	{
		return FALSE;
	}
}
#endif // LIBPNG_SUPPORT
